@@ -1,5 +1,4 @@
-"""
-Matplotlib plus async
+"""Matplotlib plus async
 
 
 mainloop()
@@ -17,7 +16,7 @@ numbers.
 
 You want to see what is going on without grinding the calculations to a halt.
 
-At the same time, the more you explore the dyata with plots (thanks matplotlib)
+At the same time, the more you explore the data with plots (thanks matplotlib)
 the more questions appear.
 
 Often the calculation has a number of parameters, or a selection of data
@@ -29,8 +28,11 @@ me to control more of the parameters.
 This module aims to provide a small number of components to help asynchronous
 data exploration and graphical monitoring of data processing pipelines.
 
-It is particularly focussed on global climate data that typically comes in a
-lat/lon grid of values.
+It also aims to help with the process of maintenance of data streams
+over time.   See metagit.py for more ideas there.
+
+Many magic classes started as objects in other short scripts, see
+examples for some of those.
 
 Internals?
 ==========
@@ -53,15 +55,43 @@ I am starting off here with some pieces from my project *karmapi*.
 
 So there will be Pig Farms, Piglets and widgets all mixed up for a while.
 
-Objects with a start and run.
+Update
+======
 
-Some writing to queues where viewers are polling.
+I am planning to re-work how objects pass between Balls.
+
+Balls await put() and get() messages and not care where they come from.
+
+The plan is for the GeeFarm or Shepherd to allow you to examine queue
+statistics and add and delete edges to the graph, with an option
+whereby it randomly connects things together until it works or melts
+or something.
+
+Make it a bit easier to view and navigate the various graphs.
+
+Leaning to a bunch of different *Controller* objects that you can call
+up for different tasks such as viewing queues, starting and stopping
+processes.
+
+Other objects, waiting on input.
+
+In short, there is a working prototype.  
+
+It is time to untangle the various objects and mold it into something
+smaller and more powerful.
+
+That we can then use in the farm module to build a functioning whole.
+
+Let the balls bounce around randomly with the magic roundabout routing
+the photons.
 
 """
 import random
 
 import math
 
+import datetime
+
 import io
 
 import sys
@@ -70,8 +100,22 @@ from pathlib import Path
 
 from collections import deque, defaultdict, Counter
 
+import argparse
+
+import operator
+
+from traceback import print_exc
+
+import inspect
+
+import dateutil
+
 import curio
 
+sleep = curio.sleep
+run = curio.run
+spawn = curio.spawn
+
 import numpy as np
 
 from PIL import Image
@@ -84,10 +128,11 @@ from matplotlib import pyplot as plt
 
 import networkx as nx
 
+Parser = argparse.ArgumentParser
 
 class Ball:
     
-    def __init__(self):
+    def __init__(self, **kwargs):
 
         self.paused = False
         self.sleep = .3
@@ -101,7 +146,6 @@ class Ball:
         self.radii.add_filter('s', self.sleepy)
         self.radii.add_filter('w', self.wakey)
         self.radii.add_filter(' ', self.toggle_pause)
-        self.radii.add_filter('i', self.interact)
 
 
     def __getattr__(self, attr):
@@ -109,11 +153,18 @@ class Ball:
         """
         return getattr(self.radii, attr)
 
-    async def interact(self):
+    def update(self, args):
+        """ Update attributes as per args 
+
+        args is expected to be from argparse.
+
+        Now it is super easy to write a short module that has a bunch of
+        command line parameters.
 
-        from pprint import pprint
-        pprint(vars(self))
         
+        """
+        for key, value in vars(args).items():
+            setattr(self, key, value)
 
     async def sleepy(self):
         """ Sleep more between updates """
@@ -136,6 +187,326 @@ class Ball:
     async def run(self):
         pass
 
+class Interact(Ball):
+
+    def __init__(self, ball):
+
+        super().__init__()
+        
+        self.history = deque()
+        self.set_ball(ball)
+
+        self.add_filter('0', self.toggle)
+        self.add_filter('1', self.add_one)
+        self.add_filter('3', self.sub_one)
+        self.add_filter('2', self.double)
+        self.add_filter('4', self.half)
+        self.add_filter('5', self.flipsign)
+        self.add_filter('8', self.rcycle)
+        self.add_filter('9', self.cycle)
+        self.add_filter('x', self.tenx)
+        self.add_filter('z', self.tenth)
+        self.add_filter('m', self.add_m)
+        self.add_filter('\177', self.shorten)
+
+        self.add_filter(' ', self.next_attr)
+        self.add_filter('\b', self.prev_attr)
+        self.add_filter('\r', self.re_interact)
+        self.add_filter('Left', self.back)
+
+        self.add_filter('i', self.interact)
+
+
+    def set_ball(self, ball):
+
+        try:
+            self.attrs = deque(sorted(vars(ball).keys()))
+        except:
+            print('oops no can interact')
+            return
+
+        self.ball = ball
+
+
+
+    async def back(self):
+        """ pop back in history """
+
+        if not self.history:
+            return
+        
+        ball = self.history.pop()
+        self.set_ball(ball)
+
+
+    async def interact(self):
+        """ Go into interactive mode 
+        
+        hmm.... worm can time
+        """
+
+        from pprint import pprint
+        for key, value in sorted(vars(self.ball).items()):
+            rep = repr(value)
+            if len(rep) > 200:
+                rep = rep[:10] + ' ... ' + rep[-10:]
+            print(key, rep, type(value))
+
+        print()
+        self.show_current()
+        print()
+
+
+    async def re_interact(self):
+        """ Recursively interact mode 
+        
+        hmm.... bigger worm can time
+
+        Idea is to interact on current attribute of whatever current is
+        """
+        
+        current = self.current()
+        obj = getattr(self.ball, current)
+
+        self.set_ball(obj)
+        self.show_current()
+
+        
+    def show_current(self):
+        
+        attr = self.current()
+        print(attr, getattr(self.ball, attr))
+
+
+    def current(self):
+        """ Return current attr """
+        return self.attrs[0]
+
+    async def next_attr(self):
+
+        self.attrs.rotate(-1)
+
+        self.show_current()
+
+    async def prev_attr(self):
+
+        self.attrs.rotate()
+        attr = self.attrs[0]
+
+        self.show_current()
+
+
+    def operate(self, op=operator.add, b=2):
+        
+        key = self.current()
+        value = getattr(self.ball, key)
+        if b is not None:
+            value = op(value, b)
+        else:
+            value = op(value)
+
+        setattr(self.ball, key, value)
+
+        print(f'{key}: {value}')
+
+    async def double(self):
+        """ i double """
+        self.operate(operator.mul, 2) 
+        
+    async def half(self):
+        """ i half """
+        self.operate(operator.mul, 1/2)
+
+    async def tenx(self):
+        """ i ten x """
+
+        self.operate(operator.mul, 10)
+        
+    async def tenth(self):
+        """ i one tenth """
+
+        self.operate(operator.mul, 1/10)
+        
+    async def add_one(self):
+        """ i one more """
+
+        self.operate(operator.add, 1)
+
+    async def add_m(self):
+        """ i one thousand more """
+
+        self.operate(operator.add, 1000)
+
+    async def sub_one(self):
+        """ i one less """
+        self.operate(operator.add, -1)
+
+    async def toggle(self):
+        """ i toggle """
+        self.operate(operator.not_, None)
+        
+    async def flipsign(self):
+        """ i flip """
+        self.operate(operator.mul, -1)
+
+    def acycle(self, howmuch=1):
+
+        key = self.current()
+        value = getattr(self.ball, key)
+        
+        try:
+            value.rotate(howmuch)
+            if value:
+                print(value[0])
+        except:  
+            print(value)
+            try:
+                if howmuch > 0:
+                    split = -1 * (1 + howmuch)
+                    value = value[:] + value[:howmuch]
+            except:
+                print('no cycle for you', howmuch)
+
+    async def cycle(self):
+        """ i cycle """
+        self.acycle()
+        
+    async def rcycle(self):
+        """ i cycle back """
+        self.acycle(-1)
+
+    async def shorten(self):
+
+        key = self.current()
+        
+        value = getattr(self.ball, key)
+        if len(value) > 1:
+            print(f'popping {value[0]}')
+            value.popleft()
+            print(f'new head {value[0]}')
+
+class Spell:
+    """ A magic spell, or cast if you like, if it works
+
+    For help processing lists of dictionaries.
+
+    Or csv files, where we have to turn strings into values.
+
+    The game is guessing the types of the columns in a csv file.
+
+    Use a magic.Spell.
+
+    
+    """
+
+    def __init__(self, cache=None):
+        """  
+        
+        Older idea, updated::
+
+        Cache size is how much rewind we get if a type changes
+
+        If None, everything gets cached.
+
+        It would make sense to coordinate this with
+        functools.lru_cache, now known as functools.cache (python 3.9).
+        
+        For small datasets this might be what you want.
+
+        """
+
+        # casts by keyword
+        self.casts = {}
+        self.date_parse = date_parse = dateutil.parser.parser().parse
+        self.upcast = {None: int, int: float, float: date_parse, date_parse: str}
+        self.fill = {None: None, int: 0, float: 0.0, date_parse: None, str: ''}
+
+        # how much data to look at to find casts
+        self.sniff = 10
+        
+        
+        # i don't think this bit is implemented yet -- see comment above
+        self.cache = deque(maxlen=cache)
+
+
+    def spell(self, data, sniff=10):
+        """ Apply casts to data
+        
+        Would like this to be dynamic, updating the casts as we go """
+
+        # short hand for:  for xx in self.cast_data(data); yield xx
+        yield from self.cast_data(data)
+        
+        
+    def find_casts(self, data, sniff=10):
+
+        sniff = sniff or self.sniff
+
+        keys = data[0].keys()
+
+        casts = self.casts
+        
+        upcast = self.upcast
+        
+        for row in data[self.sniff:]:
+            for key in keys:
+                value = row[key].strip()
+                if value:
+                    try:
+                        casts.setdefault(key, int)(value)
+                    except:
+                        casts[key] = upcast[casts[key]]
+                    
+        # look for a (first) date key - probably should looke
+        # for all dates, really we are looking for an index here
+        self.datekey = None
+        for key, cast in self.casts.items():
+            if cast is self.date_parse:
+                self.datekey = key
+                return
+
+
+    def check_casts(self, data, sniff=10):
+
+        self.find_casts(data, sniff)
+
+    def cast_data(self, data):
+
+        casts = self.casts
+
+        fill = self.fill
+
+        for row in data:
+
+            result = {}
+            for key, value in row.items():
+                if key not in casts:
+                    print(key)
+                    continue
+                
+                cast = casts[key] 
+                if not value.strip():
+                    value = fill.setdefault(cast)
+
+                result[key] = cast(value)
+            yield result
+
+    def fields(self):
+
+        return self.casts.keys()
+
+def find_date_key(record):
+
+    for key, value in record.items():
+        try:
+
+            date = dateutil.parser.parser(value)
+            return key
+        except Exception as e:
+            # guess it is not this one
+            print(e)
+            print(key, value, 'is not a date')
+
 
 class GeeFarm(Ball):
     """ A farm, for now.. 
@@ -195,7 +566,9 @@ class GeeFarm(Ball):
         self.shep.flock = self.hub
 
         print('starting shep')
-        await self.shep.start()
+        result = self.shep.start()
+        if inspect.iscoroutine(result):
+            await result
 
         # create a task which is a dog watching the shepherd
         self.superdog = await curio.spawn(canine(self.shep))
@@ -219,8 +592,9 @@ class GeeFarm(Ball):
         await self.superdog.cancel()
 
 
+modes = deque(['grey', 'white', 'black'])
 
-def fig2data(fig):
+def fig2data(fig=None, background='grey'):
     """ Convert a Matplotlib figure to a PIL image.
 
     fig: a matplotlib figure
@@ -230,19 +604,22 @@ def fig2data(fig):
 
     FIXME -- turning matplotlib figures into PIL or numpy
     """
-    facecolor = 'white'
-    facecolor = 'grey'
-    facecolor = 'black'
+    #facecolor = 
+    #facecolor = 
+    #facecolor = 'black'
+    fig = fig or plt
+    facecolor = modes[0]
     if hasattr(fig, 'get_facecolor'):
         facecolor = fig.get_facecolor()
         #print('facecolor', facecolor)
 
     # no renderer without this
     image = io.BytesIO()
-       
+
     fig.savefig(image, facecolor=facecolor, dpi=200)
     try:
         fig.close()
+        pass
     except AttributeError:
         print('cannot close figure')
 
@@ -285,10 +662,10 @@ class RoundAbout:
 
         return qq
 
-    async def put(self, value, name='stdout'):
+    async def put(self, value=None, name='stdout'):
 
         self.counts.update([('put', name)])
-        await self.select(name).put(value)
+        await self.select(name).put(value or fig2data(plt))
 
     async def get(self, name='stdin'):
 
@@ -344,6 +721,8 @@ class Shepherd(Ball):
         self.relays = {}
         self.path = [self]
 
+        self.interaction = Interact(self.path[-1])
+
         self.add_filter('h', self.help)
         self.add_filter('n', self.next)
         self.add_filter('p', self.previous)
@@ -354,10 +733,35 @@ class Shepherd(Ball):
 
         self.add_filter('T', self.status)
 
+        self.add_filter('i', self.interact)
+
         # make a little sleepy
         self.sleep *= 10
 
 
+    async def interact(self):
+        """ Go into interactive mode 
+        
+        hmm.... worm can time
+        """
+
+
+        # Now if interaction has
+        current = self.current()
+        if current is not self.interaction.ball:
+            self.interaction.set_ball(current)
+
+        if self.interaction not in self.path:
+            self.path.append(self.interaction)
+
+        await self.interaction.interact()
+
+        
+    def current(self):
+
+        return self.path[-1]
+
+
     async def status(self):
 
         print('PATH')
@@ -397,20 +801,25 @@ class Shepherd(Ball):
 
         or just send it to anything that is running and seems to care?
         """
-        print('whsitle', key, name)
+        #print('whsitle', key, name)
         
         for sheep in reversed(self.path):
             lu = sheep.radii.filters[name]
             #print('whistle', sheep, lu)
             if key in lu.keys():
-                print('sending message', key, name, sheep)
-                await lu[key]()
+                try:
+                    await lu[key]()
+                except:
+                    print_exc()
 
                 # first one gets it?
                 return True
 
         # nobody cares :(
-        print('nobody cares :(')
+        try:
+            print('nobody cares :(', key, ord(key), type(key))
+        except:
+            print(key)
         return False
 
     async def help(self, name='keys'):
@@ -437,7 +846,7 @@ class Shepherd(Ball):
         print(msg)
         await self.put(msg, 'help')
         from blume import teakhat
-        teakhat.Help(msg)
+        #teakhat.Help(msg)
 
 
     def doc_firstline(self, doc):
@@ -458,8 +867,10 @@ class Shepherd(Ball):
                 continue
                 
             print('starting', sheep)
-            # just start all the nodes
-            await sheep.start()
+            # just start all the nodes 
+            dolly = sheep.start()
+            if inspect.iscoroutine(dolly):
+                await dolly
             
             info = self.flock.nodes[sheep]
             print('info', info)
@@ -510,7 +921,7 @@ class Shepherd(Ball):
     async def next(self):
         """ Move focus to next """
         print(f'what is next?: {self.path}')
-        print(self.path[-1])
+        print(self.current())
 
     async def previous(self):
         """ Move focus to previous """
@@ -524,10 +935,7 @@ class Shepherd(Ball):
 
     async def down(self):
         """ Move focus to next node """
-        current = None
-        
-        if self.path:
-            current = self.path[-1]
+        current = self.current()
 
         succ = nx.dfs_successors(self.flock, current, 1)
 
@@ -541,7 +949,7 @@ class Shepherd(Ball):
     async def toggle_run(self, sheep=None):
         """ Toggle run status of sheep """
 
-        sheep = sheep or self.path[-1]
+        sheep = sheep or self.current()
         
         # run it if not already running
         if sheep not in self.running:
@@ -575,7 +983,7 @@ class Shepherd(Ball):
             if sheep in self.running:
                 c = 'red'
 
-            if sheep is self.path[-1]:
+            if sheep is self.current():
                 c = 'gold'
             elif sheep in self.path:
                 c= 'green'
@@ -584,9 +992,7 @@ class Shepherd(Ball):
             
         nx.draw_networkx(self.flock, node_color=colours)
 
-        print('sheput')
-        await self.put(fig2data(plt))
-        print('shepend')
+        await self.put()
         #print(self.radii.counts)
         
         #print(self.radii)
@@ -612,18 +1018,23 @@ class Shepherd(Ball):
         return f'shepherd of flock degree {len(self.flock)}'
 
 
-class Table(Ball):
+class Table:
     """ Magic table.
 
     A list of dictionaries, and ways to explore them?
     """
     
-    def __init__self(self, data=None):
-
-        super(self).__init__()
+    def __init__self(self, path=None, data=None, meta_data=None):
+        """ Turn what we are given into a table """
 
+        self.path = path or Path()
         self.data = data
-    
+        self.meta_data = meta_data
+        
+        """ now what?  look at data with a magic spell and get meta data.
+        
+        
+        """
 
 async def canine(ball):
     """ A sheep dog, something to control when it pauses and sleeps
@@ -643,6 +1054,8 @@ async def canine(ball):
     off runs.
 
     With a bit more work when building things ... self.radii time?
+
+    Update: trying to accommodate balls where run is just a function.
     
     """
 
@@ -651,12 +1064,12 @@ async def canine(ball):
     while True:
         if not ball.paused:
 
-            await ball.run()
+            result = ball.run()
+            if inspect.iscoroutine(result):
+                await result
+            
             runs += 1
 
-            if False:
-                print(f'run {runs} for {ball}')
-
         await curio.sleep(ball.sleep)
 
 
@@ -664,11 +1077,11 @@ async def relay(a, b):
 
     while True:
         value = await a.get('stdout')
-        print('relay', type(value), 'from', type(a), 'to', type(b))
+        #print('relay', type(value), 'from', type(a), 'to', type(b))
         
         await b.put(value, 'stdin')
 
-async def run():
+async def runme():
 
     farm = GeeFarm()
 
@@ -685,11 +1098,11 @@ async def run():
 
     
 def random_colour():
-
+    """ Pick a random matplotlib colormap """
     return random.choice(plt.colormaps())
     
 
 if __name__ == '__main__':
     
     
-    curio.run(run(), with_monitor=True)
+    run(runme(), with_monitor=True)
