@@ -244,7 +244,8 @@ class Cell(Rectangle):
         bbox.y1 = bbox.y0 + height
 
         # apply transform to get width and height in table coordinates.
-        l, b, w, h = bbox.inverse_transformed(self.get_data_transform()).bounds
+        l, b, w, h = bbox.transformed(self.get_data_transform().inverted()).bounds
+
         return w, h
 
     def _get_horizontal_pad(self):
@@ -262,9 +263,7 @@ class Cell(Rectangle):
         """
         Update the text properties.
 
-        Valid kwargs are
-
-        %(Text)s
+        Valid kwargs are as per `.Text`.
         """
         self._text.update(kwargs)
         self.stale = True
@@ -406,6 +405,7 @@ class Table(Artist):
 
         self._cells = {}
         self._edges = None
+
         self._autoColumns = []
         self._autoFontsize = True
         self._has_column_labels = False
@@ -441,7 +441,6 @@ class Table(Artist):
         """
         Set a Cell in a given position.
         """
-        cbook._check_isinstance(Cell, cell=cell)
         try:
             row, col = position[0], position[1]
         except Exception:
@@ -522,7 +521,7 @@ class Table(Artist):
                  if row >= start_row and col >= 0]
         bbox = Bbox.union(boxes)
 
-        return bbox.inverse_transformed(self.get_transform())
+        return bbox.transformed(self.get_transform().inverted())
 
     def contains(self, mouseevent):
         # docstring inherited
